import torch
from transformers import AutoModelForCausalLM, AutoTokenizer, StoppingCriteria, StoppingCriteriaList

class TextGenerator:
    def __init__(self, model_name, **kwargs):
        self.tokenizer = AutoTokenizer.from_pretrained(model_name, **kwargs)
        self.model = AutoModelForCausalLM.from_pretrained(model_name, device_map="auto", output_attentions=True, **kwargs)

    def generate_text(self, input_text, max_new_tokens=100):
        input_ids = self.tokenizer(input_text, return_tensors="pt").input_ids.to("cuda")
        stopping_criteria = StoppingCriteriaList([StopOnNewLineCriteria(self.tokenizer)])
        outputs = self.model.generate(input_ids, max_new_tokens=max_new_tokens, output_attentions=True, return_dict_in_generate=True, stopping_criteria=stopping_criteria)
        generated_text = self.tokenizer.decode(outputs.sequences[0], skip_special_tokens=True)
        attentions = outputs.attentions
        return generated_text, attentions, self.tokenizer, input_ids, outputs

class StopOnNewLineCriteria(StoppingCriteria):
    def __init__(self, tokenizer):
        self.tokenizer = tokenizer

    def __call__(self, input_ids, scores, **kwargs):
        # 마지막으로 생성된 토큰이 줄바꿈인지 확인
        if input_ids[0, -1] == self.tokenizer.eos_token_id or \
            self.tokenizer.decode(input_ids[0, -1]) == "\n":
            return True
        return False

class InterSentenceAttention:
    def __init__(self, generated_text, generated_sequences, attentions, tokenizer):
        self.tokenizer = tokenizer

        # 문장을 마침표 기준으로 분리
        # sentences = [sentence.strip() for sentence in generated_text.split('\n') if sentence.strip()]
        sentence_boundaries, sentences = self._find_sentence_boundaries(generated_sequences)

        # 문장 간 attention 계산
        integrated_attentions = self._integrate_attentions(attentions)
        sentence_attention_heads = self._calculate_sentence_attention(integrated_attentions, sentence_boundaries)
        sentence_attention = self._aggregate_attention(sentence_attention_heads)

        self.sentences = sentences
        self.sentence_attention_heads = sentence_attention_heads
        self.sentence_attention = sentence_attention

    
    def _integrate_attentions(self, attentions):
        num_layers = len(attentions)
        _, num_heads, prompt_seq_length, _ = attentions[0][-1].shape
        max_seq_length = attentions[-1][-1].shape[-1]

        integrated = torch.zeros((1, num_heads, max_seq_length, max_seq_length))
        integrated[0, :, :prompt_seq_length, :prompt_seq_length] = attentions[0][-1] # 0번째 레이어(프롬프트 간의 attention을 가짐)의 마지막 히든 레이어의 heads들은 그대로 붙임
        # 나머지 레이어들(생성된 토큰들과 이전 토큰들의 attention을 가짐)의 마지막 히든 레이어의 heads들을 이어붙임
        for layer in range(1, num_layers):
            integrated[0, :, prompt_seq_length+layer:prompt_seq_length+layer+1, :prompt_seq_length+layer] = attentions[layer][-1][0]
        
        return integrated # shape: (1, num_heads, max_seq_length, max_seq_length)
    
    def _aggregate_attention(self, sentence_attentions):
        # input shape: (1, num_heads, num_sentences, num_sentences)
        print(sentence_attentions.shape)
        max_attention_heads, _ = torch.max(sentence_attentions, dim=1)
        print(max_attention_heads.shape)

        return max_attention_heads.squeeze(0) # shape: (num_sentences, num_sentences)

    def _find_sentence_boundaries(self, sequences):
        boundaries = [0]
        sentences = []

        for i, token in enumerate(sequences):
            decoded = self.tokenizer.decode(token)
            if '\n' in decoded:
                sentences.append(self.tokenizer.decode(sequences[boundaries[-1]:i]))
                boundaries.append(i)

        sentences.append(self.tokenizer.decode(sequences[boundaries[-1]:len(sequences)]))
        boundaries.append(len(sequences))

        return boundaries, sentences

    def _calculate_sentence_attention(self, attentions, sentence_boundaries):
        # 문장 간 attention 계산
        # num_layers = len(attentions)
        # num_layers = 1
        _, num_heads, _, seq_length = attentions.shape # shape: (1, num_heads, seq_length, seq_length)
        num_sentences = len(sentence_boundaries) - 1 # 시작점과 끝점을 포함하므로 1 뺌
        print(sentence_boundaries)

        # 0으로 초기화
        sentence_attentions = torch.zeros((1, num_heads, num_sentences, num_sentences))

        # for layer in range(num_layers):
        # layer = 0
        for head in range(num_heads):
            for i in range(num_sentences):
                start_i = sentence_boundaries[i]
                end_i = sentence_boundaries[i + 1]
                for j in range(num_sentences):
                    start_j = sentence_boundaries[j]
                    end_j = sentence_boundaries[j + 1]
                    # 문장 범위 내 텐서 추출
                    attention_slice = attentions[0, head, start_i:end_i, start_j:end_j]
                    
                    if attention_slice.numel() == 0: # 빈 텐서 검사
                        max_attention = 0
                    else:
                        max_attention = torch.max(attention_slice).item() # 문장 내 토큰들의 attention의 최대값
                    
                    sentence_attentions[0, head, i, j] = max_attention # 각 헤드 내에서 i번째 문장관 j번째 문장의 attention

        return sentence_attentions # shape: (1, num_heads, num_sentences, num_sentences)



import matplotlib
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
# import mplcursors
import numpy as np
import textwrap
from matplotlib.colors import LinearSegmentedColormap

class InterSentenceAttentionVisualization:
    def __init__(self, sentence_attention, sentences):
        self.sentence_attention = sentence_attention
        self.sentences = sentences

    def visualize_sentence_attention(self):
        num_sentences = len(self.sentences)
        print(num_sentences)

        fig = plt.figure(figsize=(num_sentences/2+1, num_sentences/3))
        gs = gridspec.GridSpec(2, 2, height_ratios=[5, 0.2], width_ratios=[1, 1], hspace=0.3)

        ax = plt.subplot(gs[0, 0])
        # ax.set_xlim(0, num_x_sentences - 0.5)
        # ax.set_ylim(-0.5, num_y_sentences - 0.5)

        x, y = np.meshgrid(np.arange(num_sentences), np.arange(num_sentences))
        x = x.flatten()
        y = y.flatten()
        colors = self.sentence_attention.transpose(0, 1).flatten()

        # 커스텀 그라디언트 색상 맵 생성
        custom_colors = ["black", "blue", "cyan", "lime", "yellow", "orange", "red"]
        cmap = LinearSegmentedColormap.from_list("custom_gradient", custom_colors)

        # colors 배열을 넘파이 배열로 변환
        colors_np = np.array(colors)

        # 컬러바 추가
        ax_cbar = plt.subplot(gs[1, 0])
        norm = plt.Normalize(vmin=np.min(colors_np), vmax=np.max(colors_np))
        cb1 = plt.colorbar(plt.cm.ScalarMappable(norm=norm, cmap=cmap), cax=ax_cbar, orientation='horizontal')
        cb1.ax.tick_params(labelleft=False, labelright=True)
        cb1.set_label('Attention Score')

        # 0인 값은 white로, 그 외는 커스텀 그라디언트를 적용
        color_values = np.array([cmap(val) if val > 0 else (1, 1, 1, 1) for val in colors_np])

        # 모든 포인트를 포함하여 scatter 플롯 생성
        scatter = ax.scatter(x, y, color=color_values, cmap=cmap, s=150, edgecolors='none')

        ax.set_xticks(np.arange(num_sentences))
        ax.set_yticks(np.arange(num_sentences))

        plt.grid(False)

        # 텍스트 영역 추가
        text_ax = plt.subplot(gs[0, 1])
        text_ax.axis('off')  # 축 비활성화

        text = text_ax.text(0.5, 0.5, "Hover over points", ha='center', va='center', transform=text_ax.transAxes)
        
        def on_motion(event):
            cont, ind = scatter.contains(event)
            if cont:
                index = ind["ind"][0]
                index_x = index // num_sentences
                index_y = index % num_sentences
                generated_sentence = self._wrap_text(self.sentences[index_y], 50)
                focused_sentence = self._wrap_text(self.sentences[index_x], 50)
                attention = f"{self.sentence_attention[index_y, index_x]:.5f}"
                text.set_text(f"[x: Generated Sentence]\n{generated_sentence}\n\n[y: Focused Sentence]\n{focused_sentence}\n\n[Attention]\n{attention}")
                # text.set_text(f"Generated Sentence: {sentences[index_y]}\n\nFocused Sentence: {sentences[index_x]}\n\nAttention: {sentence_attention[index_y, index_x]}")
            else:
                text.set_text("Hover over points")
            fig.canvas.draw_idle()

        # 마우스 이동 이벤트에 대한 핸들러 등록
        fig.canvas.mpl_connect('motion_notify_event', on_motion)

        # mplcursors를 사용하여 호버 기능 추가
        # cursor = mplcursors.cursor(scatter, hover=True)
        # @cursor.connect("add")
        # def on_add(sel):
        #     index_x = sel.target.index // num_x_sentences
        #     index_y = sel.target.index % num_y_sentences
        #     sel.annotation.set(text=f"Focused Sentence: {sentences[index_x]}\nGenerated Sentence: {sentences[index_y]}\nAttention: {sentence_attention[index_y, index_x]}",
        #                        position=(0, 0))  # 어노테이션 위치
        #     sel.annotation.get_bbox_patch().set(fc="white", alpha=0.6)

        plt.show()

    def visualize_sentence_attention_heatmap(self):
        # 문장 간 어텐션 매트릭스의 크기
        num_sentences = self.sentence_attention.shape[0]
        
        # 히트맵 생성
        fig, ax = plt.subplots(figsize=(num_sentences, num_sentences))
        cax = ax.matshow(self.sentence_attention, cmap='viridis')
        
        # 축에 문장을 레이블로 추가
        ax.set_xticks(np.arange(num_sentences))
        ax.set_yticks(np.arange(num_sentences))
        ax.set_xticklabels(self.sentences, rotation=90)
        ax.set_yticklabels(self.sentences)
        
        # 컬러바 추가
        fig.colorbar(cax)
        
        plt.show()

    def _wrap_text(self, text, width):
        """
        텍스트가 width 사이즈 넘어갈 시 줄바꿈
        """
        return '\n'.join(textwrap.wrap(text, width))




from setuptools import setup, find_packages

setup(
    name="savis",
    version="0.1",
    license="MIT",
    packages=find_packages(),
    install_requires=[
        "transformers>=4.0.0",
        "matplotlib",
        "torch>=1.7.1"
    ],
    author="Seongbum Seo",
    author_email="seo@seongbum.com",
    description="A Sentence-Level Visualization of Attention in Transformers",
    keywords="text generation visualization transformers",
    classifiers=[
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ],
)
